// Stream.cpp

#include "stdafx.h"
#include "Stream.h"

namespace IO
{
	namespace Private
	{
		////////////////////////////////////////////////////////////
		// MemoryStream
		// nꂽ̉Ȃǂ͍sȂ
		// TODO: |V[x[X̉H
		class MemoryStream : public IStream
		{
		private:
			const BYTE	*m_pBegin, *m_pEnd, *m_pCurrent;
			volatile LONG	m_ReferenceCount;
		public:
			MemoryStream(const void* data, size_t size) : m_ReferenceCount(1)
			{
				m_pBegin = m_pCurrent = (BYTE*)data;
				m_pEnd = m_pBegin + size;
			}
		protected:
			~MemoryStream()
			{
				ASSERT(m_ReferenceCount == 0);
			}
		public:
			// Ԃ̖ɑΏ邽߂ operator new/delete قǂ̂
			// void* operator new (size_t size);
			// void operator delete (void* ptr);
			STDMETHODIMP_(ULONG) AddRef()
			{
				return ::InterlockedIncrement(&m_ReferenceCount);
			}
			STDMETHODIMP_(ULONG) Release()
			{
				ULONG refs = ::InterlockedDecrement(&m_ReferenceCount);
				if(refs == 0)
				{
					delete this;
				}
				return refs;
			}
			STDMETHODIMP QueryInterface(REFIID iid, void** ppObj)
			{
				ASSERT(*ppObj == NULL);
				if(iid == __uuidof(IStream) || iid == __uuidof(ISequentialStream) || iid == __uuidof(IUnknown))
				{
					*ppObj = static_cast<IStream*>(this);
					AddRef();
					return S_OK;
				}
				*ppObj = NULL;
				return E_NOINTERFACE;
			}
			STDMETHODIMP Read(void* buffer, ULONG size, ULONG* done)
			{
				ASSERT(m_pEnd >= m_pCurrent);
				UINT32 remain = (UINT32)(m_pEnd - m_pCurrent);
				if(size > remain) size = remain;
				memcpy(buffer, m_pCurrent, size);
				m_pCurrent += size;
				if(done) *done = size;
				return size > 0 ? S_OK : S_FALSE;
			}
			STDMETHODIMP Seek(LARGE_INTEGER move, DWORD method, ULARGE_INTEGER* newpos)
			{
				switch(method)
				{
				case STREAM_SEEK_SET:
					if(move.QuadPart < 0 || (ULONGLONG)move.QuadPart > GetSize())
						return STG_E_INVALIDPOINTER;
					m_pCurrent = m_pBegin + move.QuadPart;
					break;
				case STREAM_SEEK_CUR:
				{
					const BYTE* pos = m_pCurrent + move.QuadPart;
					if(pos < m_pBegin || pos > m_pEnd)
						return STG_E_INVALIDPOINTER;
					m_pCurrent = pos;
					break;
				}
				case STREAM_SEEK_END:
					if(move.QuadPart > 0 || (ULONGLONG)(-move.QuadPart) > GetSize())
						return STG_E_INVALIDPOINTER;
					m_pCurrent = m_pEnd + move.QuadPart;
					break;
				default:
					//TRESPASS();
					return E_INVALIDARG;
				}
				if(newpos)	newpos->QuadPart = m_pCurrent - m_pBegin;
				return S_OK;
			}
			STDMETHODIMP Clone(IStream** ppStream)
			{
				*ppStream = new MemoryStream(m_pBegin, (size_t)(m_pEnd - m_pBegin));
				return S_OK;
			}
			STDMETHODIMP Stat(STATSTG* pStat, DWORD flag)
			{
				if(flag != STATFLAG_NONAME)
				{	// Ƃ肠OȂ
					pStat->pwcsName[0] = L'\0';
				}
				else
				{	// NULL ɂĂȂƖ|C^ŎʃAvP[V
					//iGDI+PNGǂݍ񂾏ꍇȂǁGȂ݂ɁAPNGȊO͂Ȃvj
					pStat->pwcsName	= NULL;
				}
				pStat->type					= STGTY_LOCKBYTES;
				pStat->cbSize.QuadPart		= GetSize();
				pStat->grfMode				= STGM_READ | STGM_SHARE_DENY_WRITE;
				pStat->grfLocksSupported	= 0;
				pStat->clsid				= CLSID_NULL;
				pStat->grfStateBits			= 0;
				return S_OK;
			}
			// ̃\bh
			STDMETHODIMP Commit(DWORD flag)	{ return E_NOTIMPL; }
			STDMETHODIMP LockRegion(ULARGE_INTEGER offset, ULARGE_INTEGER size, DWORD locktype)		{ return E_NOTIMPL; }
			STDMETHODIMP UnlockRegion(ULARGE_INTEGER offset, ULARGE_INTEGER size, DWORD locktype)	{ return E_NOTIMPL; }
			STDMETHODIMP CopyTo(IStream* pStream, ULARGE_INTEGER size, ULARGE_INTEGER* read, ULARGE_INTEGER* written)	{ return E_NOTIMPL; }
			// ȃ\bh
			STDMETHODIMP Write(const void* buffer, ULONG size, ULONG* done)	{ return STG_E_ACCESSDENIED; }
			STDMETHODIMP SetSize(ULARGE_INTEGER size)	{ return STG_E_ACCESSDENIED; }
			STDMETHODIMP Revert()			{ return STG_E_ACCESSDENIED; }
			// 
			ULONGLONG  GetSize()	{ return (ULONGLONG)(m_pEnd - m_pBegin); }
		};
	}

	HRESULT	Stream::CreateOnFile(LPCTSTR filename, DWORD mode)
	{
		ASSERT(m_pStream == NULL);
		HRESULT hr = ::SHCreateStreamOnFile(filename, mode, &m_pStream);
		if FAILED(hr)
		{
			//TRACE("@Stream::CreateOnFile(filename=", filename, ") : HRESULT=", hr);
			return hr;
		}
		return S_OK;
	}
	HRESULT	Stream::CreateOnHGlobal(HGLOBAL hGlobal, BOOL deleteOnRelease)
	{
		ASSERT(m_pStream == NULL);
		return ::CreateStreamOnHGlobal(hGlobal, deleteOnRelease, &m_pStream);
	}
	HRESULT	Stream::CreateOnResource(HINSTANCE hInstance, LPCTSTR name, LPCTSTR type)
	{
		ASSERT(m_pStream == NULL);
		HRSRC hResInfo = ::FindResource(hInstance, name, type);
		void* data = ::LockResource(::LoadResource(hInstance, hResInfo));
		if(data == NULL)
		{
			//TRACE("@Stream::CreateOnResource(name=", name, ", type=", type, ") : Resource not found.");
			return E_FAIL;
		}
		m_pStream = new Private::MemoryStream(data, ::SizeofResource(hInstance, hResInfo));
		return S_OK;
	}
}
